# Rest Microservices with Spring Boot


## Index
* [Scenario Schema](#scenario-schema)
* [customer-service Example](#customer-service-example)
* [product-service Example](#product-service-example)
* [order-service Example](#order-service-example)
* [Communication With Microservices](#communication-with-microservices)
* [Dockerizing Microservices](#dockerizing-microservices)


---------------------------------------------------------------------------------------------------------------------

## Scenario Schema

![scenerio](./img/scenerio.jpg)

As you can see three services will be created in this lab, but the link between them will be established later.


## customer-service Example

1. Generate a new stand-alone Spring application using [Spring Initialzr](https://start.spring.io/). (Can find the setup folder)

2. Fill up the necessary fields (Group, Artifact and Dependencies) as shown below.

![spring-initializr](./img/spring-initializr.jpg)
```
Gruop: com.eteration
Artifact: customer-service
Project: maven
Dependicies: Spring Web, Config Client
```

3. Then click the generate button(ctrl+enter) project will be downloaded as a compressed file, extract it to a suitable directory

4. Launch Eclipse EE and click the import project.

5. The generated project as an Existing Maven Project as shown below.

![existing-maven](./img/existing-maven.jpg)

6. Click the browse and select the extracted file(customer-service).

![customer-import](./img/customer-import.jpg)


7. If everything is done correctly, one should have a similar folder structure as displayed below.

![customer-check](./img/customer-check.jpg)


8. So let's create **com.eteration.customerservice.dto** DTO package and **CustomerDTO.Java** class.

![customer-dto](./img/customer-dto.jpg)

* Implement the following attributes of the Customer DTO:
```
     ID (String)
     name (String)
     surname (String)
     address (String)
     cardNumber(String)
```

* Generate the getter and setter for each attribute. 

> Shortcut: **Shift+Alt+S** -> **R** -> **selected item** -> **Generate**

9. If everything is done correctly, one should have a similar class structure as displayed below.

```java
package com.eteration.customerservice.dto;

public class CustomerDTO {

	private String id;
	private String name;
	private String surname;
	private String address;
	private String cardNumber;
	

	public String getId() {
		return id;
	}
	public void setId(String id) {
		this.id = id;
	}
	public String getName() {
		return name;
	}
	public void setName(String name) {
		this.name = name;
	}
	public String getSurname() {
		return surname;
	}
	public void setSurname(String surname) {
		this.surname = surname;
	}
	public String getAddress() {
		return address;
	}
	public void setAddress(String address) {
		this.address = address;
	}
	public String getCardNumber() {
		return cardNumber;
	}
	public void setCardNumber(String cardNumber) {
		this.cardNumber = cardNumber;
	}
}
```

10. So let's create **com.eteration.customerservice.service** Service package and **CustomerService.Java** class.

![customer-service](./img/customer-service.jpg)

* Implement the method readCustomer(String id)

```java
package com.eteration.customerservice.service;

import org.springframework.stereotype.Service;
import com.eteration.customerservice.dto.CustomerDTO;

@Service
public class CustomerService {
	
	public CustomerDTO readCustomer(String id) {
		CustomerDTO customer = new CustomerDTO();
		customer.setId(id);
		customer.setAddress("İstanbul/Bahcelievler");
		customer.setName("Murat");
		customer.setSurname("Karakas");
		customer.setCardNumber("1234-4545-4545-2325");

		return customer;
	}
}
```

11. So let's create **com.eteration.customerservice.controller** Controller package and **CustomerController.Java** class.

![customer-controller](./img/customer-controller.jpg)


* Implement this annotation at head of CustomerController class.

```java 
@RestController
@RequestMapping("/customers")
public class CustomerController {
```

> Make this class a Rest Controller by adding **@RestController** annotation at class level. Add a base url path for request mapping again at class level by using **@RequestMapping** annotation.


* implement the method customerService

```java
@Autowired
private CustomerService customerService;
```

> Inject the Customer Service here by using **@Autowired** annotation.



* Implement the call service.readCustomer(id) method.

```java
@GetMapping("/{customerId}")
public CustomerDTO customerId(@PathVariable("customerId") String id) {
    return customerService.readCustomer(id);
}
```

> Map a **GET** request to readCustomer(...) method, created in Customer Service by using **@GetMapping** annotation. Indicate the specific URI that will map to this method by using **@PathVariable** annotation.


12. If everything is done correctly, one should have a similar class structure as displayed below.

```java
package com.eteration.customerservice.controller;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import com.eteration.customerservice.dto.CustomerDTO;
import com.eteration.customerservice.service.CustomerService;

@RestController
@RequestMapping("/customers")
public class CustomerController {

	@Autowired
	private CustomerService customerService;


	@GetMapping("/{customerId}")
	public CustomerDTO customerId(@PathVariable("customerId") String id) {
	    return customerService.readCustomer(id);
	}
	
}
```

* Lastly for future purposes, add a application.yaml file under src/main/resources as below, you can clear this directory if you notice other files and it’s confusing you. And Set the kdv to use the price.


```yaml
spring:
  application:
	name: customer-service
```


13. Run **CustomerServiceApplication.java**, go to the specified port on a browser, initiate a new query as shown below and check the results.

> [http://localhost:8080/customers/1](http://localhost:8080/customers/1)  let's check here.

![customer-results](./img/customer-results.jpg)


---------------------------------------------------------------------------------------------------------------------

## product-service Example 

1. Generate a new stand-alone Spring application using [Spring Initializr](https://start.spring.io/). (Easy way, can we find setup folder)

2. Fill up the necessary fields (Group, Artifact and Dependencies) as shown below.

![spring-initializr-product](./img/spring-initializr-product.jpg)

```
Gruop: com.eteration
Artifact: product-service
Project: maven
Dependicies: Spring Web, Actuator
```

3. Then click the generate button(ctrl+enter) project will be downloaded as a compressed file, extract it to a suitable directory

4. Launch Eclipse EE and click the import project.

5. The generated project as an Existing Maven Project as shown below.

![existing-maven](./img/existing-maven.jpg)

6. Click the browse and select the extracted file(product-service).

![product-import](./img/product-import.jpg)

7. If everything is done correctly, one should have a similar folder structure as displayed below.

![every-thing-okay-product](./img/every-thing-okay-product.jpg)


8. So let's create **com.eteration.productservice.dto** DTO package and **ProductDTO.Java** class.

![productdto](./img/productdto.jpg)

* Implement the following attributes of the Product DTO:
```java
private String productId;
private String name;
private double price;
```


* Generate the getter and setter for each attribute. 

> Shortcut: **Shift+Alt+S** -> **R** -> **selected item** -> **Generate**

9. If everything is done correctly, one should have a similar class structure as displayed below.

```java
package com.eteration.productservice.dto;

public class ProductDTO {
	
	private String productId;
	private String name;
	private double price;
	
	public String getProductId() {
		return productId;
	}
	public void setProductId(String productId) {
		this.productId = productId;
	}
	public String getName() {
		return name;
	}
	public void setName(String name) {
		this.name = name;
	}
	public double getPrice() {
		return price;
	}
	public void setPrice(double price) {
		this.price = price;
	}
}
```

10. So let's create **com.eteration.productservice.service** Service package and **ProductService.Java** class.

![product-service](./img/product-service.jpg)

* Implement the method readProduct(String productId)

```java
package com.eteration.productservice.service;

import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.web.bind.annotation.PathVariable;

import com.eteration.productservice.dto.ProductDTO;

@Service
public class ProductService {

	@Value("${kdv}")
	private int kdv;
	
	public ProductDTO readProduct(@PathVariable("productId") String productId) {
		ProductDTO product = new ProductDTO();
		product.setProductId(productId);
		product.setName("Product" + productId);
		product.setPrice(Math.random() * 100 * kdv);
		return product;
	}
}
```

11. So let's modified **com.eteration.productservice.controller** Controller package and **ProductController.Java** class.

```java
package com.eteration.productservice.controller;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import com.eteration.productservice.dto.ProductDTO;
import com.eteration.productservice.service.ProductService;

@RestController
@RequestMapping("/products")
public class ProductServiceController {
	
	@Autowired
	private ProductService productService;
	
	@GetMapping("/{productId}")
	public ProductDTO readProduct(@PathVariable("productId") String productId) {
		return productService.readProduct(productId);
	}
}
```


* Lastly for future purposes, add a  **application.yaml** file. under src/main/resources as below.


```yaml
spring:
  application:
    name: product-service
    	
kdv: 8
```


12. Run **ProductServiceApplication.java** as a Java Application, go to the [http://localhost:8080/products/32133](http://localhost:8080/products/32133) a browser and make the request with respect to the URI specified in the controller and check the results. Notice the product name with the suffix. 

> We can try with different id's and you will see different results.
http://localhost:8080/products/{id}

![product-result](./img/product-result.jpg)




---------------------------------------------------------------------------------------------------------------------------


## order-service Example


1. After confirming Product Service is running without error. Generate a Spring Boot Application called order-service using Spring Initialzr by following the steps before. Generated project will be downloaded as a compressed file, extract it to a suitable directory. (Easy way, can we find setup folder)

2. Click the import project.

3. The generated project as an Existing Maven Project as shown below.

![existing-maven](./img/existing-maven.jpg)

4. Click the browse and select the extracted file(order-service in setup folder).

![order-import](./img/order-import.jpg)

5. So let's create **com.eteration.orderservice.dto** DTO package and **OrderRequestDTO.Java** class.

![order-request-dto](./img/order-request-dto.jpg)

* Implement the following attributes of the Order DTO:

```java
private long customerId;
private long productId;
private int count;
```

* Generate the getter and setter for each attribute. 

> Shortcut: **Shift+Alt+S** -> **R** -> **selected item** -> **Generate**

```java
package com.eteration.orderservice.dto;

public class OrderRequestDTO {
	
	private long customerId;
	private long productId;
	private int count;
	
	public long getCustomerId() {
		return customerId;
	}
	public void setCustomerId(long customerId) {
		this.customerId = customerId;
	}
	public long getProductId() {
		return productId;
	}
	public void setProductId(long productId) {
		this.productId = productId;
	}
	public int getCount() {
		return count;
	}
	public void setCount(int count) {
		this.count = count;
	}	
}
```

6. Create anathor Class in **com.eteration.orderservice.dto** DTO package and **OrderResponseDTO.Java** class.

![order-response-dto](./img/order-response-dto.jpg)


* Implement the following attributes of the Order DTO:

```java
private String orderId;
private String message;
private double totalAmount;
```

* Generate the getter and setter for each attribute. 

> Shortcut: **Shift+Alt+S** -> **R** -> **selected item** -> **Generate**

```java
package com.eteration.orderservice.dto;

public class OrderResponseDTO {

	private String orderId;
	private String message;
	private double totalAmount;
	
	public String getOrderId() {
		return orderId;
	}
	public void setOrderId(String orderId) {
		this.orderId = orderId;
	}
	public String getMessage() {
		return message;
	}
	public void setMessage(String message) {
		this.message = message;
	}
	public double getTotalAmount() {
		return totalAmount;
	}
	public void setTotalAmount(double totalAmount) {
		this.totalAmount = totalAmount;
	}
}
```

7. So let's create **com.eteration.orderservice.service** service package and **OrderService.Java** class.

![order-service](./img/order-service.jpg)

* Implement the method createOrder(@RequestBody OrderRequestDTO orderRequest)

```java
package com.eteration.orderservice.service;

import java.util.UUID;
import org.springframework.stereotype.Service;
import org.springframework.web.bind.annotation.RequestBody;
import com.eteration.orderservice.dto.OrderRequestDTO;
import com.eteration.orderservice.dto.OrderResponseDTO;

@Service
public class OrderService {

	public OrderResponseDTO createOrder(OrderRequestDTO orderRequest)
	{

		OrderResponseDTO orderResponseDTO = new OrderResponseDTO();
		orderResponseDTO.setOrderId(UUID.randomUUID().toString());
		orderResponseDTO.setMessage("The order that contains (" +
				orderRequest.getCount() + ") products with productID=" +
				orderRequest.getProductId() +
				" belonging to customer with customerID=" +
				orderRequest.getCustomerId()+ " is being prepared !");

		orderResponseDTO.setTotalAmount(Math.random() * 150 + 50);

		return orderResponseDTO;
	}
	
}
```


8. So let's create **com.eteration.orderservice.controller** DTO package and **OrderController.Java** class.

![order-controller](./img/order-controller.jpg)

* Modify the content of this class as below.

```java
package com.eteration.orderservice.controller;


import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import com.eteration.orderservice.dto.OrderRequestDTO;
import com.eteration.orderservice.dto.OrderResponseDTO;
import com.eteration.orderservice.service.OrderService;

@RestController
@RequestMapping("/orders")
public class OrderController {
	
	@Autowired
	private OrderService orderService;
	
	@PostMapping
	public OrderResponseDTO createOrder(@RequestBody OrderRequestDTO orderRequest) {
		return orderService.createOrder(orderRequest);
	}
}
```


> This controller will make a **POST** request whose body contains the necessary information provided by an orderRequest parameter (of type OrderRequestDTO).

* Use **@PostMapping** annotation at the method level to indicate it is a **POST** request mapping and indicate this parameter will be passed in request body by using @RequestBody annotation.

> This controller will return an **orderResponse** (of type OrderResponseDTO).

* Lastly for future purposes, add a application.yaml file under src/main/resources as below, you can clear this directory if you notice other files and it’s confusing you.

```yaml
spring:
  application:
    name: order-service
```

9. Open a tool that is able to make HTTP requests [Postman](https://www.postman.com/downloads/) after running **OrderServiceApplication.java**, start constructing the response body as below and 

- Follow the red circles 
![postman](./img/postman.jpg)

* make the request:
```json
{
"customerId": 4,
"productId": 2,
"count":"3"
}
```

* excpected output like this:
```json
{
    "orderId": "fb86800b-7c53-40ae-93c4-6d545098b681",
    "message": "The order that contains (3) products with productID=2 belonging to customer with customerID=4 is being prepared !",
    "totalAmount": 100.47299433241841
}
```

## Communication With Microservices

* **Developing clients in order-service**

1. In order to run all services in our machine, we should set their ports differently for each other. 

* For customer-services **application.yaml** file under src/main/resources as below config
```
spring:
  application:
    name: customer-service
    
server:
   port: 8081
```

* For product-services **application.yaml** file under src/main/resources as below config
```
spring:
  application:
    name: product-service
    
server:
   port: 8082
```

* For order-services **application.yaml** file under src/main/resources as below config
```
spring:
  application:
    name: order-service
    
server:
   port: 8080
```

2. Create a package named Client in order-service/src/main/java 

![order-client-package](./img/order-client-package.jpg)

* After that create a class named ProductClientService in order-service/src/main/java/

```java
package com.eteration.orderservice.client;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.web.client.RestTemplate;

@Component
public class ProductClientService {
	
	@Autowired
	private RestTemplate restTemplate;
	
	public ProductDTO readProductById(long productId) {
		return restTemplate.getForObject("http://localhost:8082/products/"+productId, ProductDTO.class);
	}

}

```

* We should create a ProductDTO, write the following code 
![order-productDTO](./img/order-productDTO.jpg)

``` java
package com.eteration.orderservice.dto;

public class ProductDTO {
	
	private String productId;
	private String name;
	private double price;
	
	public String getProductId() {
		return productId;
	}
	public void setProductId(String productId) {
		this.productId = productId;
	}
	public String getName() {
		return name;
	}
	public void setName(String name) {
		this.name = name;
	}
	public double getPrice() {
		return price;
	}
	public void setPrice(double price) {
		this.price = price;
	}
}
```
* After that return back to the ProductClientService and import ProductDTO to prevent compile issue. 


* Let's do the same thing for customer-service. So create CustomerServiceClient in order-service/src/main/java 

``` java
package com.eteration.orderservice.client;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.web.client.RestTemplate;

import com.eteration.orderservice.dto.CustomerDTO;

@Component
public class CustomerClientService {
	
	@Autowired
	private RestTemplate restTemplate;
	
	public CustomerDTO readCustomerById(long customerId) {
		return restTemplate.getForObject("http://localhost:8081/customers/"+customerId, CustomerDTO.class);
	}

}

```

*  We should create a CustomerDTO, write the following code 
![order-customerDTO](./img/order-customerDTO.jpg)

``` java
package com.eteration.orderservice.dto;

public class CustomerDTO {

	private String id;
	private String name;
	private String surname;
	private String address;
	private String cardNumber;
	

	public String getId() {
		return id;
	}
	public void setId(String id) {
		this.id = id;
	}
	public String getName() {
		return name;
	}
	public void setName(String name) {
		this.name = name;
	}
	public String getSurname() {
		return surname;
	}
	public void setSurname(String surname) {
		this.surname = surname;
	}
	public String getAddress() {
		return address;
	}
	public void setAddress(String address) {
		this.address = address;
	}
	public String getCardNumber() {
		return cardNumber;
	}
	public void setCardNumber(String cardNumber) {
		this.cardNumber = cardNumber;
	}
}
```

* In order to inject RestTemplate Bean in our client classes, we have to define the code below into Spring Main classes for each services (order-service, product-service and customer-service)

``` java
	@Bean
	public RestTemplate restTemplate(RestTemplateBuilder builder) {
	// Do any additional configuration here
	return builder.build();
	}
```

* Let's replace hard-coded code block from order-serevice with new client services.

``` java
package com.eteration.orderservice.service;

import java.util.UUID;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.web.bind.annotation.RequestBody;

import com.eteration.orderservice.client.CustomerClientService;
import com.eteration.orderservice.client.ProductClientService;
import com.eteration.orderservice.dto.CustomerDTO;
import com.eteration.orderservice.dto.OrderRequestDTO;
import com.eteration.orderservice.dto.OrderResponseDTO;
import com.eteration.orderservice.dto.ProductDTO;

@Service
public class OrderService {
	
	@Autowired
	public CustomerClientService customerClientService;

	@Autowired
	public ProductClientService productClientService;
	
	public OrderResponseDTO createOrder(@RequestBody OrderRequestDTO orderRequest)
	{
		OrderResponseDTO orderResponseDTO = new OrderResponseDTO();
		
		CustomerDTO customerDTO = customerClientService.readCustomerById(orderRequest.getCustomerId());
		
		ProductDTO productDTO = productClientService.readProductById(orderRequest.getProductId());

		orderResponseDTO.setMessage("Dear customer "+customerDTO.getName() +" "+customerDTO.getSurname() +
		" your order "+productDTO.getName()+" is send to your address: "+customerDTO.getAddress());

		
		orderResponseDTO.setOrderId(UUID.randomUUID().toString());


		orderResponseDTO.setTotalAmount(productDTO.getPrice() * orderRequest.getCount() );

		return orderResponseDTO;
	}
	
}
```

* Let's check, is it working?
``` json
{
    "customerId": 1,
    "productId": 1,
    "count": 7
}
```
![postman-order-post](./img/postman-order-post.jpg)




* **Configuring clients**

 We have defined constants for client url's and it should not be 'localhost' for every environment of the application. We make our service's configurable by environment variables to gain more flexibility. Let's make it configurable from application.yaml file. 

 ``` yaml
spring:
  application:
    name: order-service
    
server:
   port: ${PORT:8080}
   
client:
   product-service: 
      url: ${PRODUCT_PATH:http://localhost:8082}
   customer-service:
      url: ${CUSTOMER_PATH:http://localhost:8081}
 ```

 * So let's make port number configurable from application.yaml file for customer-services under src/main/resources as below config 
```yaml
spring:
  application:
    name: customer-service

server:
   port: ${PORT:8081}
```

* Also let's make port number configurable from application.yaml file for product-services under src/main/resources as below config
```yaml
spring:
  application:
    name: product-service

kdv: 8

server:
   port: ${PORT:8082}
```

  * After that we should get a url from application.yaml in our client services. To achive that, we will use `@Value` annotation of the Spring Boot.  

 ```java
package com.eteration.orderservice.client;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.web.client.RestTemplate;

import com.eteration.orderservice.dto.CustomerDTO;

@Component
public class CustomerClientService {
	
	@Autowired
	private RestTemplate restTemplate;
	
	@Value("${client.customer-service.url}")
	private String baseUrl;
	
	public CustomerDTO readCustomerById(long customerId) {
		return restTemplate.getForObject(baseUrl + "/customers/"+customerId, CustomerDTO.class);
	}

}
 ```

 ```java
package com.eteration.orderservice.client;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.web.client.RestTemplate;

import com.eteration.orderservice.dto.ProductDTO;

@Component
public class ProductClientService {
	
	@Autowired
	private RestTemplate restTemplate;
	
	@Value("${client.product-service.url}")
	private String baseUrl;
	
	public ProductDTO readProductById(long productId) {
		return restTemplate.getForObject(baseUrl + "/products/"+productId, ProductDTO.class);
	}

}
```

## Dockerizing Microservices

**Before you start:** Create a docker network.

```
docker network create microservices
```

1. Create a file named **Dockerfile** in product-service's root folder. 

2. So let's continue, copy these lines on Dockerfile

```dockerfile
FROM openjdk:11-jre-slim
RUN apt-get update
RUN apt-get install wget -y
ADD target/product-service-*.jar app.jar
ENTRYPOINT ["java","-jar","app.jar"]
```

After the creating Dockerfile file, let's take a docker image. Run the commands below in product-service root folder
```
mvn package
docker build -t product-service .
```

If every thing is okay, you will see the similar output.

![product-dockerize](./img/product-dockerize.jpg)

Let's run the docker image.
```
docker run --network microservices -e PORT=8080  --name product-service -p 8080 product-service

```
![product-docker-run](./img/product-docker-run.jpg)

3. Create a file named **Dockerfile** in customer-service's root folder. 

4. So let's continue, copy these lines on Dockerfile

```dockerfile
FROM openjdk:11-jre-slim
RUN apt-get update
RUN apt-get install wget -y
ADD target/customer-service-*.jar app.jar
ENTRYPOINT ["java","-jar","app.jar"]
```

After the creating Dockerfile file, let's take a docker image. Run the commands below in customer-service root folder

```
mvn package
docker build -t customer-service .
```

If every thing is okay, you will see the similar output.

![customer-dockerize](./img/customer-dockerize.jpg)

Let's run the docker image 
```
docker run --network microservices -e PORT=8080  --name customer-service -p 8080 customer-service
```
![customer-docker-run](./img/customer-docker-run.jpg)

5. Create a file named **Dockerfile** in order-service's root folder. 

6. So let's continue, copy these lines on Dockerfile

```dockerfile
FROM openjdk:11-jre-slim
RUN apt-get update
RUN apt-get install wget -y
ADD target/order-service-*.jar app.jar
ENTRYPOINT ["java","-jar","app.jar"]
```

After the creating Dockerfile file, let's take a docker image. Run the commands below in order-service root folder
```
mvn package
docker build -t order-service .
```

If every thing is okay, you will see the similar output.

![order-dockerize](./img/order-dockerize.jpg)

Let's run the docker image 
```
docker run --name order-service --network microservices -p 8080:8080 -e PRODUCT_PATH=http://product-service:8080 -e CUSTOMER_PATH=http://customer-service:8080 -e PORT=8080 order-service
```
![order-docker-run](./img/order-docker-run.jpg)

* We have used environment variables to configure services properly. We run 3 services but we use only `8080` port of the host machine. Currently `order-service` is available from the `localhost` but `customer-service` and `product-service` are only available from the `microservices` network because in our case, there is no need to access them from outside of the Docker environment. It is enough to make them accessable from the `order-service` and we achived that with creating `microservices` network and attaching all services to it. Now open postman and send the request below;


![postman-docker-discovery](./img/postman-docker-discovery.jpg)


## Clean up
```
 docker rm product-service -f
 docker rm customer-service -f
 docker rm order-service -f
 docker network rm microservices
```
